/*
	PSP VSH 24bpp text bliter
*/
#include <psptypes.h>
#include <pspkernel.h>
#include <pspsdk.h>
#include <pspsysclib.h>

#include "common.h"

#define ALPHA_BLEND		1
#define PB_CHAR_WIDTH			8
#define PB_CHAR_HEIGHT			8
#define	VRAM_SIZE				0x00200000
#define	VRAM_ADDR_TOP			0x40000000

#define PB_TRANSPARENT			0


extern unsigned char msx[];

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
typedef struct
{
	int				width;
	int				height;

	void		*buffer;

	int				format;
	int				lineWidth;
	
	unsigned int	frameSize;
	unsigned int	lineSize;
	unsigned char	pixelSize;
} vram_frame_buffer;

vram_frame_buffer vram = {0};

u32 fcolor = 0xFFffffff;
u32 bcolor = 0x00000000;

/////////////////////////////////////////////////////////////////////////////


#define	PB_INVALID_FONT			('?' * PB_CHAR_HEIGHT)
#define	PB_INVALID_SJIS_INDEX	(0xFFFF)
//#define PB_BLEND_FUNC( src_ch, dst_ch, alpha, maxbits ) ( ( ( ( src_ch ) + 1 - ( dst_ch ) ) * ( alpha ) >> ( maxbits ) ) + ( dst_ch ) )

/*
#define PB_SWAP( a, b )		\
{							\
	int t = *(int *)a;		\
	*(int *)a = *(int *)b;	\
	*(int *)b = t;			\
}
*/
/*
#define PB_PUT_PIXEL( adr, color )														\
{																						\
	unsigned int pc = color;															\
	if( vram.colorConv ) pc = ( vram.colorConv )( pc, (void *)(vram.addr.buffer+adr) );	\
	memcpy( (void *)(vram.addr.buffer+adr), (void *)&pc, vram.pixelSize );				\
}
*/

#define libmConvert8888_5650( src )	((((src) & 0x00f80000) >> 8) | (((src) & 0x0000fc00) >> 5) | (((src) & 0x000000f8) >> 3))
#define	libmConvert8888_5551( src )	((( ( (src) >> 24 ) ? 1 : 0 ) << 15) | (((src) & 0x00f80000) >> 9 ) | (((src) & 0x0000f800) >> 6 ) | (((src) & 0x000000f8) >> 3 ))
#define	libmConvert8888_4444( src )	((((src) & 0xf0000000) >> 16) | (((src) & 0x00f00000) >> 12) | (((src) & 0x0000f000) >> 8 ) | (((src) & 0x000000f0) >> 4 ))

#define libmConvert4444_8888( src )	((((src) & 0xf000) << 28) | (((src) & 0xf000) << 24) | (((src) & 0x0f00) << 20) | (((src) & 0x0f00) << 16) | (((src) & 0x00f0) << 12) | (((src) & 0x00f0) << 8 ) | (((src) & 0x000f) << 4 ) | (((src) & 0x000f) << 0 ))
#define libmConvert5551_8888( src )	((((src) & 0x8000) << 16) | (((src) & 0x8000) << 15) | (((src) & 0x8000) << 14) | (((src) & 0x8000) << 13) | (((src) & 0x8000) << 12) | (((src) & 0x8000) << 11) | (((src) & 0x8000) << 10) | (((src) & 0x8000) << 9) | (((src) & 0x7c00) << 9) | (((src) & 0x7000) << 4) | (((src) & 0x03e0) << 6) | (((src) & 0x0380) << 1) | (((src) & 0x001f) << 3) | (((src) & 0x001c) >> 2))
#define libmConvert5650_8888( src )	(0xff000000 | (((src) & 0xf800) << 8) | (((src) & 0xe000) << 3) | (((src) & 0x07e0) << 5) | (((src) & 0x0600) >> 1) | (((src) & 0x001f) << 3) | (((src) & 0x001c) >> 2))


#define	CONVERT_COLOR(c_s,c_d,format )															\
{																								\
	if(format == PSP_DISPLAY_PIXEL_FORMAT_4444 )		c_d = libmConvert8888_4444(c_s);		\
	else if(format == PSP_DISPLAY_PIXEL_FORMAT_5551 ) c_d = libmConvert8888_5551(c_s);			\
	else if(format == PSP_DISPLAY_PIXEL_FORMAT_565 )	c_d = libmConvert8888_5650(c_s);		\
}


#define BLEND_FUNC( src_ch, dst_ch, alpha, maxbits ) ( ( ( ( src_ch ) + 1 - ( dst_ch ) ) * ( alpha ) >> ( maxbits ) ) + ( dst_ch ) )

static inline u32 libmAlphaBlend8888( u8 alpha, u32 src, u32 dst )
{
	if( ! alpha )
	{
		return dst;
	}
	else if( alpha == 0xFF )
	{
		return src;
	}
	else
	{
		u32 color = alpha << 24;
		color |= BLEND_FUNC( ( src       ) & 0xFF, ( dst       ) & 0xFF, alpha, 8 );
		color |= BLEND_FUNC( ( src >>  8 ) & 0xFF, ( dst >>  8 ) & 0xFF, alpha, 8 ) << 8;
		color |= BLEND_FUNC( ( src >> 16 ) & 0xFF, ( dst >> 16 ) & 0xFF, alpha, 8 ) << 16;
		return color;
	}
}

static inline u32 libmAlphaBlend4444( u8 alpha, u32 src, u32 dst )
{
	alpha = (alpha >> 4) & 0x0F;
	
	if( ! alpha )
	{
		return dst;
	}
	else if( alpha == 0xF )
	{
		return src;
	}
	else
	{
		u16 color = alpha << 12;
		
		color |= BLEND_FUNC( ( src      ) & 0xF, ( dst      ) & 0xF, alpha, 4 );
		color |= BLEND_FUNC( ( src >> 4 ) & 0xF, ( dst >> 4 ) & 0xF, alpha, 4 ) << 4;
		color |= BLEND_FUNC( ( src >> 8 ) & 0xF, ( dst >> 8 ) & 0xF, alpha, 4 ) << 8;
		
		return color;
	}
}

static inline u32 libmAlphaBlend5551( u8 alpha, u32 src, u32 dst )
{
	alpha = (alpha >> 3) & 0x1F;
	
	if( ! alpha )
	{
		return dst;
	}
	else if( alpha == 0x1F )
	{
		return src;
	}
	else
	{
		u16 color = ( alpha ? 1 : 0 ) << 15;
		
		color |= BLEND_FUNC( ( src       ) & 0x1F, ( dst       ) & 0x1F, alpha, 5 );
		color |= BLEND_FUNC( ( src >>  5 ) & 0x1F, ( dst >>  5 ) & 0x1F, alpha, 5 ) << 5;
		color |= BLEND_FUNC( ( src >> 10 ) & 0x1F, ( dst >> 10 ) & 0x1F, alpha, 5 ) << 10;
		
		return color;
	}
}

static inline u32 libmAlphaBlend5650( u8 alpha, u32 src, u32 dst )
{
	alpha = (alpha >> 2) & 0x3F;
	
	if( ! alpha )
	{
		return dst;
	}
	else if( alpha == 0x3F )
	{
		return src;
	}
	else
	{
		u16 color = 0;
		
		color |= BLEND_FUNC( ( src       ) & 0x1F, ( dst       ) & 0x1F, ( alpha >> 1 ) & 0x1F, 5 );
		color |= BLEND_FUNC( ( src >>  5 ) & 0x3F, ( dst >>  5 ) & 0x3F, alpha                , 6 ) << 5;
		color |= BLEND_FUNC( ( src >> 11 ) & 0x1F, ( dst >> 11 ) & 0x1F, ( alpha >> 1 ) & 0x1F, 5 ) << 11;
		
		return color;
	}
}

#define	ABLEND_COLOR( c_a,c_s,c_adr,c_d, format)								\
{																				\
	{																			\
		if(format == PSP_DISPLAY_PIXEL_FORMAT_8888 )							\
		{																		\
			c_d = libmAlphaBlend8888(  (c_a >> 24),c_s,  *(u32*)(c_adr) );		\
		}																		\
		else if(format == PSP_DISPLAY_PIXEL_FORMAT_4444 )						\
		{																		\
			c_d = libmAlphaBlend4444(  (c_a >> 8),c_s,  *(u16*)(c_adr));		\
		}																		\
		else if(format == PSP_DISPLAY_PIXEL_FORMAT_5551 )						\
		{																		\
			c_d = libmAlphaBlend5551(  (c_a >> 24),c_s,  *(u16*)(c_adr));		\
		}																		\
		else if(format == PSP_DISPLAY_PIXEL_FORMAT_565 )						\
		{																		\
			c_d = libmAlphaBlend5650(  (c_a >> 24),c_s,  *(u16*)(c_adr));		\
		}																		\
	}																			\
}

/*
static __inline void *pb_buf_addr( int x, int y )
{
	return (void *)( ( x + y * vram.lineWidth ) * vram.pixelSize );
}
*/
#define	MakeDrawAddr(x,y)	(void*)( vram.buffer + (( (x) + (y) * vram.lineWidth ) * vram.pixelSize ) )

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
//int blit_setup(int sx,int sy,const char *msg,int fg_col,int bg_col)
int blit_setup(void)
{
	int unk;
	
	sceDisplayGetMode( &unk, &vram.width, &vram.height);
	sceDisplayGetFrameBuf( &vram.buffer, &vram.lineWidth, &vram.format, PSP_DISPLAY_SETBUF_NEXTFRAME );
	if( vram.buffer && vram.lineWidth )
	{
//		printf("vram.buffer 0x%08X \n", (u32)vram.buffer );
//		if( (u32)vram.buffer < VRAM_SIZE )
		{
//			vram.buffer= (void *)( (u32)vram.buffer | 0x40000000 );
			vram.buffer= (void *)( (u32)vram.buffer | VRAM_ADDR_TOP );
		}
		vram.pixelSize	= ( vram.format == PSP_DISPLAY_PIXEL_FORMAT_8888 ? 4 : 2 );
		vram.lineSize	= vram.lineWidth * vram.pixelSize;
		vram.frameSize	= vram.lineSize * vram.height;
		return 0;
	}
	return -1;
}

static inline void libmPointEx(void *adr, u32 src, int format)
{
	if( format == PSP_DISPLAY_PIXEL_FORMAT_8888 )
	{
		*(u32*)adr = src;
	}
	else
	{
		*(u16*)adr = src;
	}
}

/////////////////////////////////////////////////////////////////////////////
// blit text
/////////////////////////////////////////////////////////////////////////////
void blit_set_color(int fg_col,int bg_col)
{
	fcolor = fg_col;
	bcolor = bg_col;
}

/////////////////////////////////////////////////////////////////////////////
// blit text
/////////////////////////////////////////////////////////////////////////////
int blit_string(int sx,int sy,const char *msg)
{
//	unsigned int   i;
	const char		*glyph;
	unsigned char	glyph_line_data;
	unsigned short	glyph_x, glyph_y;
	unsigned int	color;
	u32 fg_col,bg_col;
	u32 ofs, put_addr, start_addr;
	int		format = vram.format, char_width;

	if( !(vram.buffer && vram.lineWidth)) return -1;

	fg_col = (fcolor);
	bg_col = (bcolor);

	CONVERT_COLOR( fcolor, fg_col, format);
	CONVERT_COLOR( bcolor, bg_col, format);

	//VRAM̊JnAhX擾
//	start_addr      = (uintptr_t)pb_buf_addr( sx, sy );
	start_addr = (u32)MakeDrawAddr(sx,sy);
	char_width = vram.pixelSize * PB_CHAR_WIDTH;

	for(;msg[0]; msg++ )
	{
		//AhX = R[h * i 1*8 = 8bytej
		if( (unsigned char)msg[0] <= 0x88 )
		{
			glyph = (void *)&(msx[(unsigned char)msg[0] * PB_CHAR_HEIGHT]);
		}
		else
		{
			//YȂ => ?
			continue;
		}

		put_addr = start_addr;
		
		for( glyph_y = 0; glyph_y < PB_CHAR_HEIGHT; glyph_y++, put_addr += vram.lineSize )
		{
			glyph_line_data = glyph[glyph_y];
			for( glyph_x = 0, ofs = (u32)put_addr; glyph_x < PB_CHAR_WIDTH; glyph_x++, glyph_line_data <<= 1, ofs += vram.pixelSize )
			{
				color = glyph_line_data & 0x80 ? fg_col : bg_col;

#if	ALPHA_BLEND
				u32 color2 = glyph_line_data  & 0x80 ? fcolor : bcolor;
				ABLEND_COLOR( color2,color,(ofs),color, format);
#endif

//				if( color != PB_TRANSPARENT ) PB_PUT_PIXEL( put_addr, color );
				if( color )	libmPointEx((void*)(ofs) , color, format );
			}
		}
		
		start_addr += char_width;
	}

	return sx;
}

/*
int blit_string_ctr(int sy,const char *msg)
{
	int sx = (480/2 - 0x10)-strlen(msg)*(8/2);
	return blit_string(sx,sy,msg);
}
*/