/*
	PSP VSH
*/
#include <pspkernel.h>
#include <pspusb.h>
#include <pspusbstor.h>

#include <psppower.h>

#include "common.h"

extern int psp_model;

static const char * disable_enable[] ={
	"Disable",
	"Enable"
};

const char *str_default = "Default";
char str_buffer[256];

int LoadStartModule(const char *module)
{
	SceUID mod = sceKernelLoadModule(module, 0, NULL);
	if (mod >= 0 )
	{
		sceKernelStartModule(mod, strlen(module)+1 , (void *)module, NULL, NULL);
	}

	return mod;
}

int StopUnloadModule(SceUID id)
{
	SceUID r = sceKernelStopModule(id ,0 ,NULL , NULL, NULL);
	if(r < 0) return r;
	return sceKernelUnloadModule(id);
}

/////////////////////////////////////////////////////////////////////////////
// change CPU clock
/////////////////////////////////////////////////////////////////////////////

static void ( *scePowerSetClockFrequency_k)() = NULL;

u32 scePowerGetCpuClockFrequencyInt_b[4];
static int ( *scePowerGetCpuClockFrequencyInt_k)() = (void *)scePowerGetCpuClockFrequencyInt_b;

u32 scePowerGetBusClockFrequencyInt_b[4];
static int ( *scePowerGetBusClockFrequencyInt_k)() = (void *)scePowerGetBusClockFrequencyInt_b;

const u16 cpu_list[]={0, 20, 50, 75, 100, 111, 122, 133, 166, 200, 222, 266, 288, 300, 333};

#define MAX_CLK_LIST (sizeof(cpu_list)/sizeof(u16))
 
void init_clock()
{
	scePowerSetClockFrequency_k = (void *)FindProc("scePower_Service", "scePower", 0x737486F2);
	memcpy( scePowerGetCpuClockFrequencyInt_b, (void *)FindProc("scePower_Service", "scePower", 0xFDB5BFE9 ), sizeof(int) * 4);
	memcpy( scePowerGetBusClockFrequencyInt_b, (void *)FindProc("scePower_Service", "scePower", 0xBD681969 ), sizeof(int) * 4);
}

static int cpu2no(int cpu)
{
	int i;

	for(i=0;i< MAX_CLK_LIST;i++)
	{
		if(cpu==cpu_list[i])
			return i;
	
	}
	return 0;
}

void *change_game_clock(int type, int dir )
{
	int sel;
	int cpu = scePowerGetCpuClockFrequencyInt_k();
	int bus = scePowerGetBusClockFrequencyInt_k();

	if( type == FUNC_GET_STR )
	{
		u32 value = ((u32 *)scePowerSetClockFrequency_k)[1];
		sprintf(str_buffer ,( ( value == 0x00001021 ) || (value == 0 ) )?"%d/%d[LOCK]":"%d/%d",  cpu , bus );
		return str_buffer;
	}
	else
	{
		sel = cpu2no( cpu );
		if( sel == 0)
		{
			cpu ++;
			sel = cpu2no( cpu );

			if( sel != 0)
			{
				bus = cpu/2;
			}
			else
			{
				cpu--;
			}
		}	
		
		// select new
		sel = limit2(sel+dir,1,MAX_CLK_LIST-1);

		if( sel)
		{
			cpu = cpu_list[sel];
			bus = (cpu_list[sel])/2;
		}
		else
		{
			cpu = 222;
			bus = 111;
		}

		scePowerSetClockFrequency_k(cpu, cpu, bus);
//		scePowerSetClockFrequency( cpu, cpu, bus);
	}

	return NULL;
}

///////////////////////////////////////////////////////////
typedef struct
{
	const char *path;
	int uid;
}UsbModuleList;

static UsbModuleList usb_list[] = {
	{"flash0:/kd/semawm.prx"		, -1 },
	{"flash0:/kd/usbstor.prx"		, -1 },
	{"flash0:/kd/usbstormgr.prx"	, -1 },
	{"flash0:/kd/usbstorms.prx"		, -1 },
	{"flash0:/kd/usbstoreflash.prx"	, -1 },
	{"flash0:/kd/usbstorboot.prx"	, -1 },
};

#define USB_MODULE_CNT	(sizeof(usb_list)/sizeof(UsbModuleList))
void LoadUsbModules()
{
	int i;
	for(i=0;i<USB_MODULE_CNT;i++)
	{
		usb_list[i].uid = LoadStartModule( usb_list[i].path );
	}
}

void UnoadUsbModules()
{
	int i;
	for(i= (USB_MODULE_CNT - 1); 0 <= i ;i--)
	{
		if( usb_list[i].uid >= 0)
		{
			StopUnloadModule(usb_list[i].uid);
			usb_list[i].uid = -1;
		}
	}
}
void *connect_usb(int type, int dir )
{
	static int connect_flag = 0;

	if( type == FUNC_GET_STR )
	{
		return (void *)disable_enable[connect_flag];
	}
	else
	{
		if( connect_flag )
		{
			sceUsbDeactivate(0);
			sceUsbStop(PSP_USBSTOR_DRIVERNAME, 0, 0);	
			sceUsbStop(PSP_USBBUS_DRIVERNAME, 0, 0);

			UnoadUsbModules();
		}
		else
		{
			LoadUsbModules();

			sceUsbStart(PSP_USBBUS_DRIVERNAME, 0, 0);
			sceUsbStart(PSP_USBSTOR_DRIVERNAME, 0, 0);
//			sceUsbstorBootSetCapacity(0x800000);
			sceUsbActivate( (psp_model) ? 0x2D2:0x1C8 );
		}

		// select new
		connect_flag = 1- connect_flag;
	}

	return NULL;
}

/////////////////////////////////////////////////////////////
void request_shutdown()
{
	scePowerRequestStandby();
}

void request_suspend()
{
	scePowerRequestSuspend();
}

///////////////////////////////////////////////////////////////
char file_path[256];
char *execute_path;

void init_gamepath()
{
	char *initfile = sceKernelInitFileName();
	strcpy( file_path, initfile);

	u32* ( *InitForKernel_040C934B )() = (void *)FindProc( "sceModuleManager", "InitForKernel", 0x040C934B );
	execute_path = (void *)((InitForKernel_040C934B())[2]);
}


static char *get_key(int api)
{
	char *ret = NULL;
	switch( api)
	{
	case 0x121:
	case 0x140:
	case 0x151:
		ret = "updater";
		break;
	case 0x120:
	case 0x141:
	case 0x142:
	case 0x152:
	case 0x153:
	case 0x160:
		ret = "game";
		break;
	case 0x123:
	case 0x124:
	case 0x125:
	case 0x126:
		ret = "umdemu";
		break;
	case 0x143:
	case 0x154:
		ret = "app";
		break;
	case 0x144:
	case 0x155:
		ret = "pops";
		break;
	case 0x145:
	case 0x156:
		ret = "licensegame";
		break;
	case 0x170:
	case 0x171:
		ret = "mlnapp";
		break;
	}

	return ret;
}

static void reset_elf(const char *path)
{
	const char *filename = path;
	int file_len = strlen(filename) + 1;

	SceUID h_block = sceKernelAllocPartitionMemory(2, "user_block", 0, sizeof(struct SceKernelLoadExecParam) + file_len , NULL);
	struct SceKernelLoadExecParam *param = (void *)sceKernelGetBlockHeadAddr(h_block);
	memset( param, 0, sizeof(struct SceKernelLoadExecParam) );

	param->size = sizeof(struct SceKernelLoadExecParam);
	param->argp = ((char *)param) + sizeof(struct SceKernelLoadExecParam);
	param->args = file_len;
	strcpy( param->argp, filename);
	filename = param->argp;

	int (* sceKernelLoadExec_k)() = (void *)FindProc("sceLoadExec", "LoadExecForUser", 0xBD2F1094 );

	int k1 = pspSdkSetK1( 0x00100000 );
	sceKernelLoadExec_k( filename, param );
	pspSdkSetK1(k1);

//	printf("ret 0x%08X \n", ret );
	sceKernelFreePartitionMemory( h_block );
}

static void FixPath(char *path)
{
	char *p = strstr( path, "EBOOT.BIN");
	if( p)
	{
		strcpy( p, "BOOT.BIN");
	}
}

void request_reset_device()
{
	int api = sceKernelInitApitype();

//	printf("api = %d : path = %s path2 = %s \n", api, file_buff, execute_path);

	if( execute_path == NULL )
		execute_path = file_path;

	if( api < 0x120 )
	{
		FixPath( file_path );
		reset_elf( file_path );
	}
	else
	{
		struct SceKernelLoadExecVSHParam param;

		memset(&param, 0, sizeof(param));
		param.size = sizeof(param);
		param.key =  get_key(api);
		param.argp = file_path;
		param.args = strlen(file_path) + 1;

		sctrlKernelLoadExecVSHWithApitype( api , execute_path , &param);
	}
}

///////////////////////////////////////////////////////////////
void request_reset_vsh()
{
	sctrlKernelExitVSH( NULL );
}

///////////////////////////////////////////////////////////////
#define MENU_RESUME		0
#define MENU_SUSPEND	1
#define MF_MENU_MAX_NUMBER_OF_THREADS 64

int setThreadsStat( int stat )
{
	static int thid_list_Current[MF_MENU_MAX_NUMBER_OF_THREADS];
	static int thid_count_Current;

	int i;
	int (*func)(SceUID) =  ( stat == MENU_SUSPEND ? sceKernelSuspendThread : sceKernelResumeThread );
	SceKernelThreadInfo thinfo;
	
	if( stat == MENU_SUSPEND )
	{
		if( sceKernelGetThreadmanIdList( SCE_KERNEL_TMID_Thread, thid_list_Current, MF_MENU_MAX_NUMBER_OF_THREADS, &thid_count_Current ) < 0 )
		{
			return -1;
		}

		for( i = 0; i < thid_count_Current; i++ )
		{
			memset(&thinfo, 0, sizeof(SceKernelThreadInfo));
			thinfo.size = sizeof(SceKernelThreadInfo);
			sceKernelReferThreadStatus(thid_list_Current[i], &thinfo);
			
			if(( thinfo.attr & 0xFF0000FF ) != 0x800000FF )
			{
//				printf("%s \n", thinfo.name );
				thid_list_Current[i] = 0;
				continue;
			}
		}
	}

	for( i = 0; i < thid_count_Current; i++ )
	{
		if( thid_list_Current[i] ) ( func )( thid_list_Current[i] );
	}

	return 0;
}

void request_pause()
{
	do
	{
		sceKernelDelayThread( 1000);
	}while( get_kernel_key() & ALL_KEYS);

	sceDisplayWaitVblankStart();
	setThreadsStat( MENU_SUSPEND );
	sceDisplayWaitVblankStart();
	sceDisplayWaitVblankStart();
	sceDisplayWaitVblankStart();
	sceDisplayWaitVblankStart();

	if( blit_setup() >= 0)
	{
		blit_set_color( 0xffffffff , 0xffff8080);
		blit_string( 0 , 0 ," Press any key to return" );
	}	

	do
	{
		sceKernelDelayThread( 10*1000);
	}while( ( get_kernel_key() & ALL_KEYS) == 0 );

	setThreadsStat( MENU_RESUME );
}
