/*
 *
 */
#include <pspkernel.h>
#include <pspsysmem_kernel.h>
#include <stdio.h>

#include <psploadexec_kernel.h>

#include "common.h"


int TSRThread(SceSize args, void *argp);

PSP_MODULE_INFO("GameSatelite", 0x1000, 1, 0);

extern int menu_sel;
extern char menu_title[];

extern u32 title_color, menu_color, cursor_color;

int menu_mode  = 0;
u32 cur_buttons = 0;//0xFFFFFFFF;
static u32 button_on  = 0;
int stop_flag=0;

int psp_model = 0;

u32 open_key = PSP_CTRL_SELECT | PSP_CTRL_START;//PSP_CTRL_VOLUP|PSP_CTRL_VOLDOWN;


const Menu_pack main_menu[] =
{
	{"CPU CLOCK      "	, TMENU_FUNCTION_VALUE	, change_game_clock		, 0 },
	{"USB CONNECT    "	, TMENU_FUNCTION_VALUE	, connect_usb			, 0 },
	{"SHUTDOWN DEVICE"	, TMENU_FUNCTION_EXIT	, request_shutdown		, CENTER(sizeof("SHUTDOWN DEVICE"))	},
	{"SUSPEND DEVICE "	, TMENU_FUNCTION_EXIT	, request_suspend		, CENTER(sizeof("SHUTDOWN DEVICE"))	},
	{"PAUSE GAME"		, TMENU_FUNCTION_EXIT	, request_pause			, CENTER(sizeof("PAUSE GAME"))		},
	{"RESET GAME"		, TMENU_FUNCTION_EXIT	, request_reset_device	, CENTER(sizeof("PAUSE GAME"))		},
	{"EXIT GAME"		, TMENU_FUNCTION_EXIT	, request_reset_vsh		, CENTER(sizeof("EXIT GAME"))		},
	{"EXIT"				, TMENU_EXIT			, NULL					, CENTER(sizeof("EXIT"))			},
	{NULL,0,"", 0}

};

#define TMENU_MAX	(sizeof(main_menu)/sizeof(Menu_pack))

int main_menu_switch[TMENU_MAX];

int thread_id = -1;

static void CreateThread()
{
	int	thid = sceKernelCreateThread("GameMenu_Thread", TSRThread, 24 , 0x1000 ,0 ,0);

	thread_id = thid;

	if (thid>=0)
	{
		sceKernelStartThread(thid, 0, 0);
	}
}

u32 get_kernel_key()
{
	/*
	SceCtrlData kernel_pad;
//	sceCtrlReadBufferPositive( &kernel_pad, 1);
	sceCtrlPeekBufferPositive( &kernel_pad, 1);
	return kernel_pad.Buttons & (
			PSP_CTRL_SELECT|PSP_CTRL_START|
			PSP_CTRL_UP|PSP_CTRL_RIGHT|PSP_CTRL_DOWN|PSP_CTRL_LEFT|
			PSP_CTRL_LTRIGGER|PSP_CTRL_RTRIGGER|
			PSP_CTRL_TRIANGLE|PSP_CTRL_CIRCLE|PSP_CTRL_CROSS|PSP_CTRL_SQUARE|
			PSP_CTRL_HOME|PSP_CTRL_NOTE);
	*/		
	return cur_buttons;
}



u32 get_user_key()
{
	u32 ret = button_on;
	button_on = 0;
	return ret;
}

int KeyThread(SceSize args, void *argp)
{
	u32 buttons;
	SceCtrlData kernel_pad;

	while(1)
	{
//		sceCtrlReadBufferPositive( &kernel_pad, 1);
		sceCtrlPeekBufferPositive( &kernel_pad, 1);
		buttons		= kernel_pad.Buttons & (
			PSP_CTRL_SELECT|PSP_CTRL_START|
			PSP_CTRL_UP|PSP_CTRL_RIGHT|PSP_CTRL_DOWN|PSP_CTRL_LEFT|
			PSP_CTRL_LTRIGGER|PSP_CTRL_RTRIGGER|
			PSP_CTRL_TRIANGLE|PSP_CTRL_CIRCLE|PSP_CTRL_CROSS|PSP_CTRL_SQUARE|
			PSP_CTRL_HOME|PSP_CTRL_NOTE|PSP_CTRL_SCREEN|
			PSP_CTRL_VOLUP|PSP_CTRL_VOLDOWN);

		button_on   = ~cur_buttons & buttons;
		cur_buttons = buttons;

		sceKernelDelayThread( 50*1000);
	}

	return 0;
}

/////////////////////////////////////////////////////////////////////////////
void EatKey(SceCtrlData *pad_data, int count, int negative)
{
	int i;
		
	if( pad_data )
	{
		int k1 = pspSdkSetK1(0);
		if( thread_id >= 0)
		{
			for(i=0;i < count;i++)
			{
				if( negative == 0 )
					pad_data[i].Buttons &= ~ALL_KEYS;
				else
					pad_data[i].Buttons |= ALL_KEYS;
			}
		}
		else
		{
			if( (cur_buttons & open_key) == open_key )
			{
				for(i=0;i < count;i++)
				{
					if( negative == 0 )
						pad_data[i].Buttons &= ~ALL_CTRL;
					else
						pad_data[i].Buttons |= ALL_KEYS;
				}
				CreateThread();
			}
		}
		pspSdkSetK1(k1);
	}
}

int sceCtrlReadBufferPositivePatched(SceCtrlData *pad_data, int count)
{
	int result = sceCtrlReadBufferPositive(pad_data,count);
	EatKey(pad_data,count, 0);
	return result;
}

int sceCtrlPeekBufferPositivePatched(SceCtrlData *pad_data, int count)
{
	int result = sceCtrlPeekBufferPositive(pad_data,count);
	EatKey(pad_data,count, 0);
	return result;
}

int sceCtrlPeekBufferNegativePatched(SceCtrlData *pad_data, int count)
{
	int result = sceCtrlPeekBufferNegative(pad_data,count);
	EatKey(pad_data,count, 1);
	return result;
}

int sceCtrlReadBufferNegativePatched(SceCtrlData *pad_data, int count)
{
	int result = sceCtrlReadBufferNegative(pad_data,count);
	EatKey(pad_data,count, 1);
	return result;
}

////////////////////////////////////////////////////////////
static void get_plugin_dir(const char *argp, char *plugin_dir )
{
	strcpy( plugin_dir , argp );	
	char *p = strrchr( plugin_dir ,'/');
	p[1] = 0;
}

int module_start(int argc, char *argp )
{
	if ( sceKernelInitKeyConfig() == PSP_INIT_KEYCONFIG_VSH )
		return 1;

	init_gamepath();

	sctrlHENPatchSyscall(FindProc("sceController_Service", "sceCtrl", 0x1F803938 ), sceCtrlReadBufferPositivePatched);
	sctrlHENPatchSyscall(FindProc("sceController_Service", "sceCtrl", 0x3A622550 ), sceCtrlPeekBufferPositivePatched);
	sctrlHENPatchSyscall(FindProc("sceController_Service", "sceCtrl", 0x60B81F86 ), sceCtrlReadBufferNegativePatched);
	sctrlHENPatchSyscall(FindProc("sceController_Service", "sceCtrl", 0xC152080A ), sceCtrlPeekBufferNegativePatched);

	psp_model = sceKernelGetModel();
	init_clock();

	char plugin_dir[256];
	get_plugin_dir( argp, plugin_dir );
	load_config( plugin_dir );
	open_key     = get_config("KEY_CONFIG");
	title_color  = get_config("TITLE_BACK_COLOR");
	menu_color   = get_config("MENU_BACK_COLOR");
	cursor_color = get_config("CURSOR_BACK_COLOR");

	u32 devkit_version = sceKernelDevkitVersion();
	menu_title[0] = ( devkit_version >> 24 ) + '0';	
	menu_title[2] = ((devkit_version >> 16 ) & 0xFF ) + '0';
	menu_title[3] = ((devkit_version >> 8  ) & 0xFF ) + '0';

	ClearCaches();

	int	thid = sceKernelCreateThread("KeyThread", KeyThread, 32 , 0x1000 ,0 ,0);
	if (thid>=0)
	{
		sceKernelStartThread(thid, 0, 0);
	}

	return 0;
}

int module_stop(int argc, char *argv[])
{
	SceUInt time = 100*1000;

	stop_flag=1;

	int i = sceKernelWaitThreadEnd( thread_id , &time );

	if(i<0)
	{
		sceKernelTerminateDeleteThread(thread_id);
	}

	return 0;
}

/////////////////////////////////////////////////////////////////////////////
static void button_func( const Menu_pack *main_menu, int menu_cnt )
{
	int res;
	// menu controll
	switch(menu_mode)
	{
	case 0:	
		if( (cur_buttons & ALL_CTRL) == 0)
		{
			menu_mode = 1;
		}
		break;
	case 1:
		if( ( res = menu_ctrl( main_menu, menu_cnt ) ) != 0)
		{
			if( res == 1 )
				menu_mode = 2;
			else
				stop_flag = 1;
		}
		break;
	case 2: // exit waiting 
		// exit menu
		if( (cur_buttons & ALL_CTRL) == 0)
		{
			stop_flag = 1;
		}
		break;
	}
}

/////////////////////////////////////////////////////////////////////////////
int TSRThread(SceSize args, void *argp)
{
//	SceCtrlData kernel_pad;

	menu_mode = 0;
	stop_flag = 0;
	menu_sel = 0;

	int menu_max_cnt = 0;
	while( main_menu[ menu_max_cnt ].path != NULL)
	{
		menu_max_cnt++;
	}

	while( stop_flag == 0 )
	{
		if(menu_mode > 0)
		{
			// check & setup video mode		
			if( blit_setup() >= 0)
			{
				menu_draw( main_menu ,menu_max_cnt );

			}

			if( sceDisplayWaitVblankStart() < 0)
				break;
		}

		sceKernelDelayThread( 800 );
//		sceKernelDelayThread( 1000 );
		button_func( main_menu, menu_max_cnt );
	}

	do
	{
		sceKernelDelayThread( 10*1000);
	}while( ( get_kernel_key() & ALL_KEYS ) != 0 );

	thread_id = -1;
	sceKernelExitDeleteThread(0);
	return 0;
}
